import logging
import os
import sys
from unittest import mock
from unittest.mock import Mock
from unittest.mock import patch

import skosprovider.skos
from pyramid.paster import get_appsettings
from pyramid.request import Request
from skosprovider.exceptions import ProviderUnavailableException
from skosprovider.providers import DictionaryProvider
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from sqlalchemy.orm import sessionmaker
from webtest import TestApp

from atramhasis import main
from atramhasis.cache import list_region
from atramhasis.cache import tree_region
from atramhasis.data.models import Provider
from atramhasis.protected_resources import ProtectedResourceEvent
from atramhasis.protected_resources import ProtectedResourceException
from fixtures.data import chestnut
from fixtures.data import larch
from fixtures.data import species
from tests import DbTest
from tests import SETTINGS
from tests import fill_db
from tests import setup_db


def setUpModule():
    setup_db()
    fill_db()


logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

here = os.path.dirname(__file__)
settings = get_appsettings(os.path.join(here, '../', 'tests/conf_test.ini'))

json_value = {
    'type': 'concept',
    'broader': [],
    'narrower': [],
    'related': [],
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'The Larch'},
        {'type': 'sortLabel', 'language': 'en', 'label': 'a'},
    ],
    'notes': [],
    'sources': [
        {
            'citation': 'Python, M.: Episode Three: How to recognise different types of trees from quite a long way away.'
        }
    ],
}

json_value_relations = {
    'broader': [{'id': '12'}],
    'id': '13',
    'related': [],
    'type': 'concept',
    'labels': [{'label': 'koperlegeringen', 'language': 'nl', 'type': 'prefLabel'}],
    'label': 'koperlegeringen',
    'notes': [],
    'narrower': [{'id': '15'}, {'id': '14'}],
}

json_value_invalid = """{
    "type": "concept",
    "broader": [],
    "narrower": [],
    "related"[]: [],
    "labels": [
        {
            "type": "prefLabel",
            "language": "en",
            "label": "The Larch"
        }
    ],
    "notes": []}
}"""

json_collection_value = {
    'labels': [{'language': 'nl', 'label': 'Test verzameling', 'type': 'prefLabel'}],
    'type': 'collection',
    'label': 'Test verzameling',
    'members': [{'id': '333'}, {'id': '7'}],
    'notes': [{'note': 'een notitie', 'type': 'note', 'language': 'nl'}],
    'infer_concept_relations': True,
}

TEST = DictionaryProvider(
    {'id': 'TEST', 'default_language': 'nl', 'subject': ['biology']},
    [larch, chestnut, species],
    concept_scheme=skosprovider.skos.ConceptScheme('http://id.trees.org'),
)


class FunctionalTests(DbTest):
    testapp: TestApp

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.init_app()

    @classmethod
    def init_app(cls):
        app = main({}, **SETTINGS)
        cls.testapp = TestApp(app)

        # Commit at end of every request. This will trigger listeners.
        class CommittingRequest(Request):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.add_finished_callback(lambda req: req.db.commit())

        cls.testapp.app.request_factory = CommittingRequest
        # change db maker to make sessions bound to the test transaction
        registry = cls.testapp.app.registry
        registry.dbmaker = sessionmaker(bind=cls.connection)

    def setUp(self):
        super().setUp()
        self.testapp.reset()

    @staticmethod
    def mock_event_handler(event):
        if event.uri == 'urn:x-vioe:geography:9':
            referenced_in = ['urn:someobject', 'https://test.test.org/object/2']
            raise ProtectedResourceException(
                f'resource {event.uri} is still in use, preventing operation',
                referenced_in,
            )

    @staticmethod
    def mock_event_handler_provider_unavailable(event):
        if event.uri == 'urn:x-vioe:geography:55':
            raise ProviderUnavailableException('test msg')


class HtmlFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_get_home(self):
        res = self.testapp.get('/', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('text/html', res.headers['Content-Type'])

    def test_get_unknown_conceptscheme(self):
        res = self.testapp.get(
            '/conceptschemes/does-not-exist',
            headers=self._get_default_headers(),
            status=404,
        )
        self.assertIn('niet gevonden', res.text)

    def test_exception_page(self):
        with mock.patch(
            'atramhasis.views.views.get_public_conceptschemes',
            side_effect=Exception('Test exception'),
        ):
            res = self.testapp.get(
                '/conceptschemes/does-not-exist',
                headers=self._get_default_headers(),
                status=500,
            )
            self.assertIn('technisch probleem', res.text)


class CsvFunctionalTests(FunctionalTests):
    def test_get_csv(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/c.csv?type=collection&label='
        )
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn(
            'attachment;filename="atramhasis_export.csv"',
            response.headers['Content-Disposition'],
        )

    def test_unicode_csv(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/c.csv?label=Chestnut&_LOCALE_=fr'
        )
        data = response.body.decode('utf-8')
        self.assertIsInstance(data, str)
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn(
            'attachment;filename="atramhasis_export.csv"',
            response.headers['Content-Disposition'],
        )
        self.assertIn('la châtaigne', data)

    def test_get_csv_all(self):
        response = self.testapp.get('/conceptschemes/TREES/c.csv')
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/csv', response.headers['Content-Type'])
        self.assertIn(
            'attachment;filename="atramhasis_export.csv"',
            response.headers['Content-Disposition'],
        )


class RestFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_get_concept(self):
        res = self.testapp.get(
            '/conceptschemes/TREES/c/1', headers=self._get_default_headers()
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['id'], '1')
        self.assertEqual(res.json['type'], 'concept')
        self.assertIn('sortLabel', [label['type'] for label in res.json['labels']])

    def test_get_conceptscheme(self):
        res = self.testapp.get(
            '/conceptschemes/TREES', headers=self._get_default_headers()
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])

    def test_get_concept_dictprovider(self):
        res = self.testapp.get(
            '/conceptschemes/TEST/c/1', headers=self._get_default_headers()
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'concept')

    def test_get_concept_not_found(self):
        res = self.testapp.get(
            '/conceptschemes/TREES/c/89',
            headers=self._get_default_headers(),
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_get_concept_dictprovider_not_found(self):
        res = self.testapp.get(
            '/conceptschemes/TEST/c/89',
            headers=self._get_default_headers(),
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_add_concept(self):
        res = self.testapp.post_json(
            '/conceptschemes/TREES/c',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'concept')

    def test_add_update_concept_manual_id(self):
        json_value['id'] = 'manual-3'
        json_value['sources'][0]['citation'] = 'short'
        res = self.testapp.post_json(
            '/conceptschemes/manual-ids/c',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual(201, res.status_code)
        res_json = res.json
        self.assertDictEqual(
            {
                'id': 'manual-3',
                'type': 'concept',
                'uri': 'urn:x-skosprovider:manual-ids:manual-3',
                'label': 'The Larch',
                'concept_scheme': {'uri': 'urn:x-vioe:manual', 'labels': []},
                'labels': [
                    {'label': 'The Larch', 'type': 'prefLabel', 'language': 'en'},
                    {'label': 'a', 'type': 'sortLabel', 'language': 'en'},
                ],
                'notes': [],
                'sources': [{'citation': 'short', 'markup': None}],
                'narrower': [],
                'broader': [],
                'related': [],
                'member_of': [],
                'subordinate_arrays': [],
                'matches': {
                    'close': [],
                    'exact': [],
                    'related': [],
                    'broad': [],
                    'narrow': [],
                },
            },
            res_json,
        )
        res_json['labels'][0]['label'] = 'updated'
        res = self.testapp.put_json(
            '/conceptschemes/manual-ids/c/manual-3',
            headers=self._get_default_headers(),
            params=res_json,
        )
        self.assertEqual(200, res.status_code)
        self.assertEqual('updated', res.json['label'])

    def test_add_concept_empty_conceptscheme(self):
        res = self.testapp.post_json(
            '/conceptschemes/STYLES/c',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])

    def test_add_concept_invalid_json(self):
        res = self.testapp.post_json(
            '/conceptschemes/TREES/c',
            headers=self._get_default_headers(),
            params=json_value_invalid,
            status=400,
        )
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_add_concept_conceptscheme_not_found(self):
        res = self.testapp.post_json(
            '/conceptschemes/GARDENNNN/c',
            headers=self._get_default_headers(),
            params=json_value,
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_conceptscheme(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES',
            headers=self._get_default_headers(),
            params=json_collection_value,
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_conceptscheme_invalid(self):
        json_collection_value.pop('labels')
        res = self.testapp.put_json(
            '/conceptschemes/TREES',
            headers=self._get_default_headers(),
            params=json_collection_value,
            expect_errors=True,
        )
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(
            res.json,
            {
                'errors': ["labels: 'labels' is a required property"],
                'message': 'Request was not valid for schema.',
            },
        )
        json_collection_value['labels'] = [
            {'language': 'nl', 'label': 'Test verzameling', 'type': 'prefLabel'}
        ]

    def test_edit_concept(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES/c/1',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_edit_concept_has_relations(self):
        res = self.testapp.put_json(
            '/conceptschemes/MATERIALS/c/13',
            headers=self._get_default_headers(),
            params=json_value_relations,
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertEqual(2, len(res.json['narrower']))

    def test_edit_concept_not_found(self):
        res = self.testapp.put_json(
            '/conceptschemes/TREES/c/89',
            headers=self._get_default_headers(),
            params=json_value,
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_delete_concept(self):
        self.testapp.delete(
            '/conceptschemes/TREES/c/1',
            headers=self._get_default_headers(),
            status=204,
        )

    def test_delete_concept_not_found(self):
        res = self.testapp.delete(
            '/conceptschemes/TREES/c/7895',
            headers=self._get_default_headers(),
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', res.status)

    def test_add_collection(self):
        res = self.testapp.post_json(
            '/conceptschemes/GEOGRAPHY/c',
            headers=self._get_default_headers(),
            params=json_collection_value,
            expect_errors=True,
        )
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'collection')

    def test_edit_collection(self):
        json_collection_value['members'] = [{'id': '7'}, {'id': '8'}]
        json_collection_value['infer_concept_relations'] = False
        res = self.testapp.put_json(
            '/conceptschemes/GEOGRAPHY/c/333',
            headers=self._get_default_headers(),
            params=json_collection_value,
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['type'], 'collection')
        self.assertEqual(2, len(res.json['members']))
        self.assertFalse(res.json['infer_concept_relations'])

    def test_delete_collection(self):
        self.testapp.delete(
            '/conceptschemes/GEOGRAPHY/c/333',
            headers=self._get_default_headers(),
            status=204,
        )

    def test_uri(self):
        res = self.testapp.post_json(
            '/conceptschemes/MATERIALS/c',
            headers=self._get_default_headers(),
            params=json_value,
        )
        self.assertEqual('201 Created', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertEqual('urn:x-vioe:materials:51', res.json['uri'])

    def test_provider_unavailable_view(self):
        def raise_provider_unavailable_exception():
            raise ProviderUnavailableException('test msg')

        with patch(
            'atramhasis.views.crud.AtramhasisCrud.delete_concept',
            Mock(side_effect=raise_provider_unavailable_exception),
        ):
            res = self.testapp.delete(
                '/conceptschemes/GEOGRAPHY/c/55',
                headers=self._get_default_headers(),
                status=503,
            )
            self.assertEqual('503 Service Unavailable', res.status)
            self.assertIn('test msg', res)

    def test_get_languages(self):
        res = self.testapp.get('/languages', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_languages_sort(self):
        res = self.testapp.get(
            '/languages', headers=self._get_default_headers(), params={'sort': 'id'}
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_languages_sort_desc(self):
        res = self.testapp.get(
            '/languages', headers=self._get_default_headers(), params={'sort': '-id'}
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res)
        self.assertEqual(len(res.json), 8)

    def test_get_language(self):
        res = self.testapp.get('/languages/de', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual('German', res.json['name'])

    def test_get_language_not_found(self):
        res = self.testapp.get(
            '/languages/jos', headers=self._get_default_headers(), expect_errors=True
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {'message': 'The resource could not be found.'})

    def test_add_language(self):
        res = self.testapp.put_json(
            '/languages/af',
            headers=self._get_default_headers(),
            params={'id': 'af', 'name': 'Afrikaans'},
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Afrikaans')

    def test_add_language_non_valid(self):
        res = self.testapp.put_json(
            '/languages/flup',
            headers=self._get_default_headers(),
            params={'id': 'flup', 'name': 'flup'},
            expect_errors=True,
        )
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(
            res.json,
            {
                'errors': [
                    {
                        'id': "Invalid language tag: Unknown code 'flup', Missing language tag in 'flup'."
                    }
                ],
                'message': 'Language could not be validated',
            },
        )

    def test_add_language_non_valid_json(self):
        res = self.testapp.put_json(
            '/languages/af',
            headers=self._get_default_headers(),
            params={'test': 'flup'},
            expect_errors=True,
        )
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(
            res.json,
            {
                'errors': {'name': 'Required'},
                'message': 'Language could not be validated',
            },
        )

    def test_edit_language(self):
        res = self.testapp.put_json(
            '/languages/de',
            headers=self._get_default_headers(),
            params={'id': 'de', 'name': 'Duits'},
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Duits')

    def test_edit_language_invalid_language_tag(self):
        res = self.testapp.put_json(
            '/languages/joss',
            headers=self._get_default_headers(),
            params={'id': 'joss', 'name': 'Duits'},
            expect_errors=True,
        )
        self.assertEqual('400 Bad Request', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(
            res.json,
            {
                'errors': [
                    {
                        'id': "Invalid language tag: Unknown code 'joss', Missing language tag in 'joss'."
                    }
                ],
                'message': 'Language could not be validated',
            },
        )

    def test_edit_language_no_id(self):
        res = self.testapp.put_json(
            '/languages/de',
            headers=self._get_default_headers(),
            params={'name': 'Duits'},
        )
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json['id'])
        self.assertEqual(res.json['name'], 'Duits')

    def test_delete_language(self):
        res = self.testapp.delete('/languages/de', headers=self._get_default_headers())
        self.assertEqual('200 OK', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])

    def test_delete_language_not_found(self):
        res = self.testapp.delete(
            '/languages/jos', headers=self._get_default_headers(), expect_errors=True
        )
        self.assertEqual('404 Not Found', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(res.json, {'message': 'The resource could not be found.'})

    def test_delete_protected_resource(self):
        def mock_event_handler(event):
            if isinstance(event, ProtectedResourceEvent):
                referenced_in = ['urn:someobject', 'https://test.test.org/object/2']
                raise ProtectedResourceException(
                    'resource {} is still in use, preventing operation'.format(
                        event.uri
                    ),
                    referenced_in,
                )

        registry = self.testapp.app.registry
        with patch.object(registry, 'notify', new=Mock(side_effect=mock_event_handler)):
            res = self.testapp.delete(
                '/conceptschemes/GEOGRAPHY/c/9',
                headers=self._get_default_headers(),
                expect_errors=True,
            )
        self.assertEqual('409 Conflict', res.status)
        self.assertIn('application/json', res.headers['Content-Type'])
        self.assertIsNotNone(res.json)
        self.assertEqual(
            res.json,
            {
                'message': 'resource urn:x-vioe:geography:9 is still in use, preventing operation',
                'referenced_in': ['urn:someobject', 'https://test.test.org/object/2'],
            },
        )

    def test_get_conceptschemes(self):
        self.testapp.get(
            '/conceptschemes', headers=self._get_default_headers(), status=200
        )

    def test_create_provider_openapi_validation(self):
        response = self.testapp.post_json(
            url='/providers',
            params={'uri_pattern': 'invalid', 'subject': 'wrong'},
            headers=self._get_default_headers(),
            expect_errors=True,
        )
        self.assertEqual(
            {
                'errors': ['None: Failed to cast value to array type: wrong'],
                'message': 'Request was not valid for schema.',
            },
            response.json,
        )
        response = self.testapp.post_json(
            url='/providers',
            params={'uri_pattern': 'invalid', 'subject': ['right']},
            headers=self._get_default_headers(),
            expect_errors=True,
        )

    def test_create_minimal_provider(self):
        response = self.testapp.post_json(
            url='/providers',
            params={
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
            },
            headers=self._get_default_headers(),
            status=201,
        )
        self.assertEqual(
            {
                'id': response.json['id'],
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': None,
                'subject': [],
                'force_display_language': None,
                'metadata': {},
                'id_generation_strategy': 'NUMERIC',
                'expand_strategy': 'recurse',
            },
            response.json,
        )

    def test_create_full_provider(self):
        response = self.testapp.post_json(
            url='/providers',
            params={
                'id': 'ERFGOEDTYPES',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            headers=self._get_default_headers(),
            status=201,
        )
        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            response.json,
        )

    def test_create_full_provider_via_put(self):
        response = self.testapp.put_json(
            url='/providers/ERFGOEDTYPES',
            params={
                'id': 'ERFGOEDTYPES',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            headers=self._get_default_headers(),
            status=201,
        )
        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
                'default_language': 'NL',
                'force_display_language': 'NL',
                'subject': ['hidden'],
                'metadata': {'Info': 'Extra data about this provider'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            response.json,
        )

    def test_update_provider(self):
        conceptscheme = ConceptScheme(
            uri='https://id.erfgoed.net/thesauri/conceptschemes'
        )
        provider = Provider(
            id='ERFGOEDTYPES',
            uri_pattern='https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
            conceptscheme=conceptscheme,
            meta={},
        )
        self.session.add(provider)
        self.session.flush()

        response = self.testapp.put_json(
            url='/providers/ERFGOEDTYPES',
            params={
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/updated/%s',
                'default_language': 'NL',
                'subject': ['hidden'],
                'force_display_language': 'NL',
                'metadata': {'extra': 'test-extra'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            headers=self._get_default_headers(),
            status=200,
        )

        self.assertEqual(
            {
                'id': 'ERFGOEDTYPES',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'https://id.erfgoed.net/thesauri/conceptschemes',
                'uri_pattern': 'https://id.erfgoed.net/thesauri/updated/%s',
                'default_language': 'NL',
                'subject': ['hidden'],
                'force_display_language': 'NL',
                'metadata': {'extra': 'test-extra'},
                'id_generation_strategy': 'MANUAL',
                'expand_strategy': 'visit',
            },
            response.json,
        )

    def test_delete_provider_with_concepts(self):
        conceptscheme = ConceptScheme(
            uri='https://id.erfgoed.net/thesauri/conceptschemes'
        )
        concept = Concept(concept_id='testconceptje', conceptscheme=conceptscheme)
        provider = Provider(
            id='ERFGOEDTYPES',
            uri_pattern='https://id.erfgoed.net/thesauri/erfgoedtypes/%s',
            conceptscheme=conceptscheme,
            meta={},
        )
        self.session.add(concept)
        self.session.add(provider)
        self.session.flush()
        conceptscheme_id = conceptscheme.id

        self.session.expire_all()
        self.assertIsNotNone(self.session.get(Provider, 'ERFGOEDTYPES'))

        self.testapp.delete(
            url='/providers/ERFGOEDTYPES',
            headers=self._get_default_headers(),
            status=204,
        )
        self.session.expire_all()
        self.assertIsNone(self.session.get(Provider, 'ERFGOEDTYPES'))
        self.assertIsNone(self.session.get(ConceptScheme, conceptscheme_id))

    def test_get_providers(self):
        response = self.testapp.get(
            url='/providers', headers=self._get_default_headers(), status=200
        )
        self.assertEqual(7, len(response.json))
        response = self.testapp.get(
            url='/providers?subject=biology',
            headers=self._get_default_headers(),
            status=200,
        )
        self.assertEqual(
            [
                {
                    'conceptscheme_uri': 'http://id.trees.org',
                    'default_language': 'nl',
                    'force_display_language': None,
                    'id': 'TEST',
                    'subject': ['biology'],
                    'type': 'DictionaryProvider',
                    'uri_pattern': 'urn:x-skosprovider:%s:%s',
                    'metadata': {},
                }
            ],
            response.json,
        )

    def test_get_provider(self):
        response = self.testapp.get(
            url='/providers/GEOGRAPHY', headers=self._get_default_headers(), status=200
        )
        self.assertEqual(
            {
                'id': 'GEOGRAPHY',
                'type': 'SQLAlchemyProvider',
                'conceptscheme_uri': 'urn:x-vioe:geography',
                'uri_pattern': 'urn:x-vioe:geography:%s',
                'default_language': None,
                'subject': [],
                'force_display_language': None,
                'id_generation_strategy': 'NUMERIC',
                'metadata': {},
                'expand_strategy': 'recurse',
            },
            response.json,
        )

    def test_expand_concept(self):
        res = self.testapp.get(
            '/conceptschemes/TREES/c/1/expand',
            headers=self._get_default_headers(),
            status=200,
        )
        self.assertEqual(res.json, ['1'])


class TestCookieView(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_cookie(self):
        response = self.testapp.get(
            '/locale?language=nl', headers=self._get_default_headers()
        )
        self.assertIsNotNone(response.headers['Set-Cookie'])
        self.assertEqual(response.status, '302 Found')
        self.assertTrue((response.headers.get('Set-Cookie')).startswith('_LOCALE_=nl'))

    def test_unsupported_language(self):
        config_default_lang = settings.get('pyramid.default_locale_name')
        response = self.testapp.get(
            '/locale?language=fr', headers=self._get_default_headers()
        )
        self.assertTrue(
            (response.headers.get('Set-Cookie')).startswith(
                '_LOCALE_=' + config_default_lang
            )
        )


class JsonTreeFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_tree(self):
        response = self.testapp.get(
            '/conceptschemes/GEOGRAPHY/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual('200 OK', response.status)
        self.assertIn('application/json', response.headers['Content-Type'])
        self.assertIsNotNone(response.json)
        self.assertEqual(2, len(response.json))
        self.assertEqual('World', response.json[0]['label'])

    def test_missing_labels(self):
        response = self.testapp.get(
            '/conceptschemes/MISSING_LABEL/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual('200 OK', response.status)
        self.assertIsNotNone(response.json)
        self.assertEqual(2, len(response.json))
        self.assertEqual('label', response.json[0]['label'])
        self.assertEqual(None, response.json[1]['label'])

    def test_tree_language(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/tree?language=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            ['De Lariks', 'De Paardekastanje'],
            [child['label'] for child in response.json[0]['children']],
        )
        response = self.testapp.get(
            '/conceptschemes/TREES/tree?language=en',
            headers=self._get_default_headers(),
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            ['The Chestnut', 'The Larch'],
            [child['label'] for child in response.json[0]['children']],
        )

    def test_no_tree(self):
        response = self.testapp.get(
            '/conceptschemes/FOO/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', response.status)


class HtmlTreeFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def test_tree(self):
        response = self.testapp.get(
            '/conceptschemes/GEOGRAPHY/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual('200 OK', response.status)
        self.assertIn('text/html', response.headers['Content-Type'])

    def test_no_tree(self):
        response = self.testapp.get(
            '/conceptschemes/FOO/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
            status=404,
            expect_errors=True,
        )
        self.assertEqual('404 Not Found', response.status)


class SkosFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'text/html'}

    def _get_json_headers(self):
        return {'Accept': 'application/json'}

    def test_admin_no_skos_provider(self):
        with patch.dict(self.testapp.app.request_extensions.descriptors):
            del self.testapp.app.request_extensions.descriptors['skos_registry']
            res = self.testapp.get(
                '/admin', headers=self._get_default_headers(), status=404
            )
        self.assertIn('niet gevonden', res.text)

    def test_crud_no_skos_provider(self):
        with patch.dict(self.testapp.app.request_extensions.descriptors):
            del self.testapp.app.request_extensions.descriptors['skos_registry']
            res = self.testapp.post_json(
                '/conceptschemes/GEOGRAPHY/c',
                headers=self._get_json_headers(),
                params=json_collection_value,
                status=404,
            )
        self.assertEqual(
            {'message': 'No SKOS registry found, please check your application setup'},
            res.json,
        )

    def test_match_filter(self):
        response = self.testapp.get(
            '/conceptschemes/TREES/c', headers={'Accept': 'application/json'}
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(3, len(response.json))
        response = self.testapp.get(
            '/conceptschemes/TREES/c'
            '?match=https://id.python.org/different/types/of/trees/nr/1/the/larch',
            headers={'Accept': 'application/json'},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            [
                {
                    'id': '1',
                    'uri': 'urn:x-skosprovider:trees/1',
                    'type': 'concept',
                    'label': 'De Lariks',
                    '@context': 'http://localhost/jsonld/context/skos',
                }
            ],
            response.json,
        )


class CacheFunctionalTests(FunctionalTests):
    def _get_default_headers(self):
        return {'Accept': 'application/json'}

    def test_create_cache(self):
        # clear entire cache before start
        invalidate_cache_response = self.testapp.get('/admin/tree/invalidate')
        self.assertEqual('200 OK', invalidate_cache_response.status)

        tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual('200 OK', tree_response.status)
        self.assertIsNotNone(tree_response.json)

        cached_tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual('200 OK', cached_tree_response.status)
        self.assertIsNotNone(cached_tree_response.json)

        self.assertEqual(tree_response.json, cached_tree_response.json)

    def test_auto_invalidate_cache(self):
        tree_region.configure(
            'dogpile.cache.memory',
            expiration_time=7000,
            arguments={'cache_size': 5000},
            replace_existing_backend=True,
        )
        list_region.configure(
            'dogpile.cache.memory',
            expiration_time=7000,
            arguments={'cache_size': 5000},
            replace_existing_backend=True,
        )
        # clear entire cache before start
        invalidate_cache_response = self.testapp.get('/admin/tree/invalidate')
        self.assertEqual('200 OK', invalidate_cache_response.status)

        tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        cached_tree_response = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual(tree_response.json, cached_tree_response.json)

        self.testapp.delete(
            '/conceptschemes/MATERIALS/c/31',
            headers=self._get_default_headers(),
            status=204,
        )

        tree_response2 = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertNotEqual(tree_response.json, tree_response2.json)

        cached_tree_response2 = self.testapp.get(
            '/conceptschemes/MATERIALS/tree?_LOCALE_=nl',
            headers=self._get_default_headers(),
        )
        self.assertEqual(tree_response2.json, cached_tree_response2.json)

        tree_region.configure('dogpile.cache.null', replace_existing_backend=True)
        list_region.configure('dogpile.cache.null', replace_existing_backend=True)


class RdfFunctionalTests(FunctionalTests):
    def test_void(self):
        rdf_response = self.testapp.get('/void.ttl')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('text/turtle', rdf_response.content_type)

    def test_rdf_full_xml(self):
        rdf_response = self.testapp.get(
            '/conceptschemes/MATERIALS/c', headers={'Accept': 'application/rdf+xml'}
        )
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_full_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_full_turtle(self):
        ttl_response = self.testapp.get(
            '/conceptschemes/MATERIALS/c', headers={'Accept': 'text/turtle'}
        )
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_full_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_xml(self):
        rdf_response = self.testapp.get(
            '/conceptschemes/MATERIALS', headers={'Accept': 'application/rdf+xml'}
        )
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_conceptscheme_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_conceptscheme_turtle(self):
        ttl_response = self.testapp.get(
            '/conceptschemes/MATERIALS', headers={'Accept': 'text/turtle'}
        )
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_conceptscheme_jsonld(self):
        res = self.testapp.get(
            '/conceptschemes/MATERIALS', headers={'Accept': 'application/ld+json'}
        )
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_conceptscheme_jsonld_ext(self):
        res = self.testapp.get('/conceptschemes/MATERIALS.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_jsonld(self):
        res = self.testapp.get(
            '/conceptschemes/MATERIALS/c/1', headers={'Accept': 'application/ld+json'}
        )
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_jsonld_ext(self):
        res = self.testapp.get('/conceptschemes/MATERIALS/c/1.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_xml(self):
        rdf_response = self.testapp.get(
            '/conceptschemes/MATERIALS/c/1', headers={'Accept': 'application/rdf+xml'}
        )
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_xml_ext(self):
        rdf_response = self.testapp.get('/conceptschemes/MATERIALS/c/1.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_turtle(self):
        ttl_response = self.testapp.get(
            '/conceptschemes/MATERIALS/c/1', headers={'Accept': 'text/turtle'}
        )
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_turtle_ext(self):
        ttl_response = self.testapp.get('/conceptschemes/MATERIALS/c/1.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_jsonld_ext_manual(self):
        res = self.testapp.get('/conceptschemes/manual-ids/c/manual-1.jsonld')
        self.assertEqual('200 OK', res.status)
        self.assertEqual('application/ld+json', res.content_type)

    def test_rdf_individual_xml_ext_manual(self):
        rdf_response = self.testapp.get('/conceptschemes/manual-ids/c/manual-2.rdf')
        self.assertEqual('200 OK', rdf_response.status)
        self.assertEqual('application/rdf+xml', rdf_response.content_type)

    def test_rdf_individual_turtle_manual(self):
        ttl_response = self.testapp.get('/conceptschemes/manual-ids/c/manual-1.ttl')
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_turtle_manual_uri(self):
        ttl_response = self.testapp.get(
            '/conceptschemes/manual-ids/c/https://id.manual.org/manual/68.ttl'
        )
        self.assertEqual('200 OK', ttl_response.status)
        self.assertEqual('text/turtle', ttl_response.content_type)

    def test_rdf_individual_not_found(self):
        self.testapp.get(
            '/conceptschemes/TREES/c/test.ttl',
            headers={'Accept': 'text/turtle'},
            status=404,
        )


class ListFunctionalTests(FunctionalTests):
    def test_labeltypes_list(self):
        labeltypeslist_res = self.testapp.get('/labeltypes')
        self.assertEqual('200 OK', labeltypeslist_res.status)
        self.assertEqual('application/json', labeltypeslist_res.content_type)
        self.assertIsNotNone(labeltypeslist_res.json)
        self.assertEqual(4, len(labeltypeslist_res.json))

    def test_notetypes_list(self):
        labeltypeslist_res = self.testapp.get('/notetypes')
        self.assertEqual('200 OK', labeltypeslist_res.status)
        self.assertEqual('application/json', labeltypeslist_res.content_type)
        self.assertIsNotNone(labeltypeslist_res.json)



"""
Testdata cotaining heritage types

.. versionadded:: 0.4.4
"""


from skosprovider.providers import DictionaryProvider
from skosprovider.uri import UriPatternGenerator
from .heritagetypes_data import HERITAGETYPESDATA

heritagetypes = DictionaryProvider(
    {'id': 'HERITAGETYPE'},
    HERITAGETYPESDATA,
    uri_generator=UriPatternGenerator('https://id.erfgoed.net/thesauri/erfgoedtypes/%s')
)



"""
Testdata cotaining event types

.. versionadded:: 0.4.4
"""


from skosprovider.providers import DictionaryProvider
from skosprovider.uri import UriPatternGenerator
from .eventtypes_data import EVENTTYPESDATA

eventtypes = DictionaryProvider(
    {'id': 'EVENTTYPE'},
    EVENTTYPESDATA,
    uri_generator=UriPatternGenerator('https://id.erfgoed.net/thesauri/gebeurtenistypes/%s')
)



import unittest

from pyramid import testing
from pyramid.httpexceptions import HTTPMethodNotAllowed
from skosprovider.providers import DictionaryProvider
from skosprovider.skos import Collection as SkosCollection
from skosprovider.skos import Concept as SkosConcept
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import Match
from skosprovider_sqlalchemy.models import MatchType
from skosprovider_sqlalchemy.models import Note
from skosprovider_sqlalchemy.models import Source
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider

from atramhasis.utils import from_thing
from atramhasis.utils import internal_providers_only
from atramhasis.utils import label_sort
from atramhasis.utils import provider_is_external
from atramhasis.utils import update_last_visited_concepts


species = {
    'id': 3,
    'uri': 'http://id.trees.org/3',
    'labels': [
        {'type': 'prefLabel', 'language': 'en', 'label': 'Trees by species'},
        {'type': 'prefLabel', 'language': 'nl', 'label': 'Bomen per soort'},
    ],
    'type': 'collection',
    'members': ['1', '2'],
    'notes': [
        {
            'type': 'editorialNote',
            'language': 'en',
            'note': 'As seen in How to Recognise Different Types of Trees from Quite a Long Way Away.',
        }
    ],
}


class TestFromThing(unittest.TestCase):
    def setUp(self):
        conceptscheme = ConceptScheme()
        conceptscheme.uri = 'urn:x-atramhasis-demo'
        conceptscheme.id = 1
        self.concept = Concept()
        self.concept.id = 11
        self.concept.concept_id = 101
        self.concept.uri = 'urn:x-atramhasis-demo:TREES:101'
        self.concept.conceptscheme_id = 1
        self.concept.conceptscheme = conceptscheme

        notes = []
        note = Note(note='test note', notetype_id='example', language_id='en')
        note2 = Note(note='note def', notetype_id='definition', language_id='en')
        notes.append(note)
        notes.append(note2)
        self.concept.notes = notes

        labels = []
        label = Label(label='een label', labeltype_id='prefLabel', language_id='nl')
        label2 = Label(label='other label', labeltype_id='altLabel', language_id='en')
        label3 = Label(
            label='and some other label', labeltype_id='altLabel', language_id='en'
        )
        labels.append(label)
        labels.append(label2)
        labels.append(label3)
        self.concept.labels = labels

        sources = []
        source = Source(citation='Kinsella S. & Carlisle P. 2015: Alice.')
        sources.append(source)
        self.concept.sources = sources

        matches = []
        match1 = Match()
        match1.uri = 'urn:test'
        match1.concept = self.concept
        match1.matchtype = MatchType(name='closeMatch', description='')
        match2 = Match()
        match2.uri = 'urn:test'
        match2.concept = self.concept
        match2.matchtype = MatchType(name='closeMatch', description='')
        matches.append(match1)
        matches.append(match2)
        self.matches = matches

        self.collection = Collection()
        self.collection.id = 12
        self.collection.concept_id = 102
        self.collection.uri = 'urn:x-atramhasis-demo:TREES:102'
        self.collection.conceptscheme_id = 1
        self.collection.conceptscheme = conceptscheme

    def test_thing_to_concept(self):
        skosconcept = from_thing(self.concept)
        self.assertTrue(isinstance(skosconcept, SkosConcept))
        self.assertEqual(skosconcept.id, 101)
        self.assertEqual(len(skosconcept.labels), 3)
        self.assertEqual(len(skosconcept.notes), 2)
        self.assertEqual(len(skosconcept.sources), 1)
        self.assertEqual(skosconcept.uri, 'urn:x-atramhasis-demo:TREES:101')

    def test_thing_to_collection(self):
        skoscollection = from_thing(self.collection)
        self.assertTrue(isinstance(skoscollection, SkosCollection))
        self.assertEqual(skoscollection.id, 102)
        self.assertEqual(skoscollection.uri, 'urn:x-atramhasis-demo:TREES:102')


class DummyViewClassForTest:
    def __init__(self):
        self.provider = None
        self.dummy = None

    def all_providers(self, dummy):
        self.dummy = dummy

    @internal_providers_only
    def internal_providers(self, dummy):
        self.dummy = dummy


class TestInternalProviderOnly(unittest.TestCase):
    def setUp(self):
        self.dummy = DummyViewClassForTest()

    def test_all_providers(self):
        self.dummy.provider = DictionaryProvider(
            list=[species], metadata={'id': 'Test'}
        )
        self.dummy.all_providers('ok')
        self.assertEqual(self.dummy.dummy, 'ok')

    def test_internal_providers(self):
        self.dummy.provider = SQLAlchemyProvider(
            metadata={'id': 'Test', 'conceptscheme_id': 1}, session=None
        )
        self.dummy.internal_providers('ok')
        self.assertEqual(self.dummy.dummy, 'ok')

    def test_external_providers(self):
        self.dummy.provider = SQLAlchemyProvider(
            metadata={'id': 'Test', 'conceptscheme_id': 1, 'subject': ['external']},
            session=None,
        )
        self.assertRaises(HTTPMethodNotAllowed, self.dummy.internal_providers, 'ok')
        self.assertIsNone(self.dummy.dummy)

    def test_no_sqlalchemyprovider(self):
        self.dummy.provider = DictionaryProvider(
            list=[species], metadata={'id': 'Test'}
        )
        self.assertRaises(HTTPMethodNotAllowed, self.dummy.internal_providers, 'ok')
        self.assertIsNone(self.dummy.dummy)


class TestUpdateLastVisitedConceptsProviderOnly(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        self.request = testing.DummyRequest()
        self.request.session = {}

    def tearDown(self):
        testing.tearDown()

    def test_update_last_visited_concepts(self):
        c = Concept()
        c.id = 2
        c.labels = [Label('test', language_id='en-us')]
        update_last_visited_concepts(
            self.request, {'label': c.label(), 'url': f'https://test.test/{55}'}
        )
        c = Concept()
        c.id = 33
        c.labels = [Label('test', language_id='nl-be')]
        update_last_visited_concepts(
            self.request, {'label': c.label(), 'url': f'https://test.test/{2}'}
        )
        self.assertEqual(2, len(self.request.session['last_visited']))

    def test_update_last_visited_concepts_max(self):
        for concept_id in range(50):
            c = Concept()
            c.id = concept_id
            c.labels = [Label('test', language_id='en-us')]
            update_last_visited_concepts(
                self.request,
                {'label': c.label(), 'url': f'https://test.test/{concept_id}'},
            )
        self.assertEqual(4, len(self.request.session['last_visited']))
        last = self.request.session['last_visited'].pop()
        self.assertEqual('https://test.test/49', last['url'])

    def test_no_double_last_visited_concepts(self):
        c = Concept()
        c.id = 2
        c.labels = [Label('test', language_id='en-us')]
        update_last_visited_concepts(
            self.request, {'label': c.label(), 'url': f'https://test.test/{55}'}
        )
        update_last_visited_concepts(
            self.request, {'label': c.label(), 'url': f'https://test.test/{55}'}
        )
        c = Concept()
        c.id = 33
        c.labels = [Label('test', language_id='nl-be')]
        update_last_visited_concepts(
            self.request, {'label': c.label(), 'url': f'https://test.test/{2}'}
        )
        self.assertEqual(2, len(self.request.session['last_visited']))


class DummyConcept:
    def __init__(self, sortlabel, language='any'):
        self.sortlabel = sortlabel
        self.language = language

    def _sortkey(self, key='sortlabel', language='any'):
        # Simulate sorting by label and language
        if language == 'any':
            return self.sortlabel
        if hasattr(self, 'labels'):
            for label in self.labels:
                if label['language'] == language:
                    return label['label']
        return self.sortlabel


def test_label_sort_empty():
    assert label_sort([]) == []


def test_label_sort_basic_sorting():
    c1 = DummyConcept('banana')
    c2 = DummyConcept('apple')
    c3 = DummyConcept('cherry')
    sorted_concepts = label_sort([c1, c2, c3])
    assert [c.sortlabel for c in sorted_concepts] == [
        'apple',
        'banana',
        'cherry',
    ]


def test_label_sort_with_language():
    class DummyConceptWithLabels(DummyConcept):
        def __init__(self, labels):
            self.labels = labels

        def _sortkey(self, key='sortlabel', language='any'):
            for label in self.labels:
                if label['language'] == language:
                    return label['label']
            return self.labels[0]['label']

    c1 = DummyConceptWithLabels(
        [
            {'label': 'appel', 'language': 'nl'},
            {'label': 'apple', 'language': 'en'},
        ]
    )
    c2 = DummyConceptWithLabels(
        [
            {'label': 'banaan', 'language': 'nl'},
            {'label': 'banana', 'language': 'en'},
        ]
    )
    c3 = DummyConceptWithLabels(
        [
            {'label': 'kers', 'language': 'nl'},
            {'label': 'cherry', 'language': 'en'},
        ]
    )
    sorted_concepts = label_sort([c1, c2, c3], language='en')
    assert [c._sortkey(language='en') for c in sorted_concepts] == [
        'apple',
        'banana',
        'cherry',
    ]
    sorted_concepts_nl = label_sort([c1, c2, c3], language='nl')
    assert [c._sortkey(language='nl') for c in sorted_concepts_nl] == [
        'appel',
        'banaan',
        'kers',
    ]


class TestProviderIsExternal(unittest.TestCase):
    def test_provider_is_external_with_external_subject(self):
        provider = MockProvider({'subject': ['external', 'other']})
        assert provider_is_external(provider) is True

    def test_provider_is_external_case_insensitive(self):
        provider = MockProvider({'subject': ['EXTERNAL']})
        assert provider_is_external(provider) is True

        provider = MockProvider({'subject': ['External']})
        assert provider_is_external(provider) is True

    def test_provider_is_not_external(self):
        provider = MockProvider({'subject': ['hidden', 'other']})
        assert provider_is_external(provider) is False

    def test_provider_is_external_no_subjects(self):
        provider = MockProvider({})
        assert provider_is_external(provider) is False

    def test_provider_is_external_empty_subjects(self):
        provider = MockProvider({'subject': []})
        assert provider_is_external(provider) is False


class MockProvider:
    """Mock provider for testing."""

    def __init__(self, metadata):
        self._metadata = metadata

    def get_metadata(self):
        return self._metadata



"""
Module containing views related to the REST service.
"""

import logging
import time

import colander
import transaction
from pyramid.httpexceptions import HTTPNoContent
from pyramid.view import view_config
from pyramid.view import view_defaults
from pyramid_skosprovider.views import ProviderView
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import NoResultFound

from atramhasis import utils
from atramhasis.audit import audit
from atramhasis.cache import invalidate_scheme_cache
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.errors import ConceptNotFoundException
from atramhasis.errors import ConceptSchemeNotFoundException
from atramhasis.errors import SkosRegistryNotFoundException
from atramhasis.errors import ValidationError
from atramhasis.json_processors import provider
from atramhasis.mappers import map_concept
from atramhasis.mappers import map_conceptscheme
from atramhasis.protected_resources import protected_operation
from atramhasis.utils import from_thing
from atramhasis.utils import internal_providers_only

LOG = logging.getLogger(__name__)


@view_defaults(accept='application/json', renderer='skosrenderer_verbose', openapi=True)
class AtramhasisCrud:
    """
    This object groups CRUD REST views part of the private user interface.
    """

    def __init__(self, request):
        self.request = request
        self.skos_manager = self.request.data_managers['skos_manager']
        self.conceptscheme_manager = self.request.data_managers['conceptscheme_manager']
        self.logged_in = request.authenticated_userid
        if 'scheme_id' in self.request.matchdict:
            self.scheme_id = self.request.matchdict['scheme_id']
            if hasattr(request, 'skos_registry') and request.skos_registry is not None:
                self.skos_registry = self.request.skos_registry
            else:
                raise SkosRegistryNotFoundException()
            self.provider = self.skos_registry.get_provider(self.scheme_id)
            if not self.provider:
                raise ConceptSchemeNotFoundException(self.scheme_id)

    def _get_json_body(self):
        json_body = self.request.json_body
        if 'c_id' in self.request.matchdict and 'id' not in json_body:
            json_body['id'] = self.request.matchdict['c_id']
        return json_body

    def _validate_concept(self, json_concept, provider, validate_id_generation):
        from atramhasis.validators import (
            Concept as ConceptSchema,
            concept_schema_validator,
        )

        concept_schema = ConceptSchema(validator=concept_schema_validator).bind(
            request=self.request,
            provider=provider,
            validate_id_generation=validate_id_generation,
        )
        try:
            return concept_schema.deserialize(json_concept)
        except colander.Invalid as e:
            raise ValidationError('Concept could not be validated', e.asdict())

    def _validate_conceptscheme(self, json_conceptscheme):
        from atramhasis.validators import (
            ConceptScheme as ConceptSchemeSchema,
            conceptscheme_schema_validator,
        )

        conceptscheme_schema = ConceptSchemeSchema(
            validator=conceptscheme_schema_validator
        ).bind(request=self.request)
        try:
            return conceptscheme_schema.deserialize(json_conceptscheme)
        except colander.Invalid as e:  # pragma no cover
            # I doubt this code will ever be reached, keeping it here just in case
            raise ValidationError('Conceptscheme could not be validated', e.asdict())

    @audit
    @view_config(
        route_name='skosprovider.conceptscheme', permission='view', request_method='GET'
    )
    def get_conceptscheme(self):
        # is the same as pyramid_skosprovider but wrapped with the audit decorator
        return ProviderView(self.request).get_conceptscheme()

    @view_config(
        route_name='skosprovider.conceptscheme', permission='edit', request_method='PUT'
    )
    def edit_conceptscheme(self):
        """
        Edit an existing concept

        :raises atramhasis.errors.ValidationError: If the provided json can't be validated
        """
        validated_json_conceptscheme = self._validate_conceptscheme(
            self._get_json_body()
        )
        conceptscheme = self.conceptscheme_manager.get(self.provider.conceptscheme_id)
        conceptscheme = map_conceptscheme(conceptscheme, validated_json_conceptscheme)
        conceptscheme = self.conceptscheme_manager.save(conceptscheme)
        self.request.response.status = '200'
        return conceptscheme

    @audit
    @view_config(route_name='skosprovider.c', permission='view', request_method='GET')
    def get_concept(self):
        """
        Get an existing concept

        :raises atramhasis.errors.ConceptNotFoundException: If the concept can't be found
        """
        c_id = self.request.matchdict['c_id']
        if isinstance(self.provider, SQLAlchemyProvider):
            try:
                concept = self.skos_manager.get_thing(
                    c_id, self.provider.conceptscheme_id
                )
            except NoResultFound:
                raise ConceptNotFoundException(c_id)
        else:
            concept = self.provider.get_by_id(c_id)
            if not concept:
                raise ConceptNotFoundException(c_id)

        self.request.response.status = '200'
        return concept

    @internal_providers_only
    @view_config(
        route_name='skosprovider.conceptscheme.cs',
        permission='edit',
        request_method='POST',
    )
    def add_concept(self):
        """
        Add a new concept to a conceptscheme

        :raises atramhasis.errors.ValidationError: If the provided json can't be validated
        """
        validated_json_concept = self._validate_concept(
            self._get_json_body(), self.provider, validate_id_generation=True
        )
        exc = None
        id_generation_strategy = IDGenerationStrategy.NUMERIC
        for _ in range(5):
            try:
                if validated_json_concept['type'] == 'concept':
                    concept = Concept()
                else:
                    concept = Collection()

                id_generation_strategy = self.provider.metadata.get(
                    'atramhasis.id_generation_strategy', IDGenerationStrategy.NUMERIC
                )
                if id_generation_strategy == IDGenerationStrategy.MANUAL:
                    concept.concept_id = validated_json_concept['id']
                else:
                    concept.concept_id = self.skos_manager.get_next_cid(
                        self.provider.conceptscheme_id, id_generation_strategy
                    )

                concept.conceptscheme_id = self.provider.conceptscheme_id
                concept.uri = self.provider.uri_generator.generate(
                    id=concept.concept_id
                )
                map_concept(concept, validated_json_concept, self.skos_manager)
                concept = self.skos_manager.save(concept)
                break
            except IntegrityError as exc:
                if id_generation_strategy == IDGenerationStrategy.MANUAL:
                    # Technically the concept_id is not 100% guaranteed to be the cause
                    # of an IntegrityError, so log trace just in case.
                    LOG.exception('Integrity error')

                    concept_id = validated_json_concept['concept_id']
                    raise ValidationError(
                        'Integrity error',
                        [{'concept_id': f'{concept_id} already exists.'}],
                    )
                # There is a small chance that another concept gets added at the same
                # time. There is nothing wrong with the request, so we try again.
                transaction.abort()
                time.sleep(0.05)
        else:
            raise Exception(f'Could not save new concept due to IntegrityErrors. {exc}')

        invalidate_scheme_cache(self.scheme_id)

        self.request.response.status = '201'
        self.request.response.location = self.request.route_path(
            'skosprovider.c', scheme_id=self.scheme_id, c_id=concept.concept_id
        )
        return from_thing(concept)

    @internal_providers_only
    @view_config(route_name='skosprovider.c', permission='edit', request_method='PUT')
    def edit_concept(self):
        """
        Edit an existing concept

        :raises atramhasis.errors.ConceptNotFoundException: If the concept can't be found
        :raises atramhasis.errors.ValidationError: If the provided json can't be validated
        """
        c_id = self.request.matchdict['c_id']
        validated_json_concept = self._validate_concept(
            self._get_json_body(), self.provider, validate_id_generation=False
        )
        try:
            concept = self.skos_manager.get_thing(c_id, self.provider.conceptscheme_id)
        except NoResultFound:
            raise ConceptNotFoundException(c_id)
        concept = map_concept(concept, validated_json_concept, self.skos_manager)

        invalidate_scheme_cache(self.scheme_id)

        self.request.response.status = '200'
        return from_thing(concept)

    @internal_providers_only
    @protected_operation
    @view_config(
        route_name='skosprovider.c', permission='delete', request_method='DELETE'
    )
    def delete_concept(self):
        """
        Delete an existing concept

        :raises atramhasis.errors.ConceptNotFoundException: If the concept can't be found
        """
        c_id = self.request.matchdict['c_id']
        try:
            concept = self.skos_manager.get_thing(c_id, self.provider.conceptscheme_id)
        except NoResultFound:
            raise ConceptNotFoundException(c_id)
        self.skos_manager.delete_thing(concept)

        invalidate_scheme_cache(self.scheme_id)
        return HTTPNoContent()

    @view_config(
        route_name='atramhasis.providers',
        permission='view',
        request_method='GET',
    )
    def get_providers(self):
        query_params = self.request.openapi_validated.parameters.query
        if 'subject' in query_params:
            filters = {'subject': query_params['subject']}
        else:
            filters = {}
        return self.request.skos_registry.get_providers(**filters)

    @view_config(
        route_name='atramhasis.provider',
        permission='view',
        request_method='GET',
    )
    def get_provider(self):
        return self.request.skos_registry.get_provider(self.request.matchdict['id'])

    @view_config(
        route_name='atramhasis.providers',
        permission='add-provider',
        request_method='POST',
    )
    def add_provider(self):
        db_provider = provider.create_provider(
            json_data=self.request.openapi_validated.body,
            session=self.request.db,
            skos_registry=self.request.skos_registry,
        )
        self.request.response.status_code = 201
        return utils.db_provider_to_skosprovider(db_provider)

    @view_config(
        route_name='atramhasis.provider',
        permission='edit-provider',
        request_method='PUT',
    )
    def update_provider(self):
        try:
            db_provider = provider.update_provider(
                provider_id=self.request.matchdict['id'],
                json_data=self.request.openapi_validated.body,
                session=self.request.db,
            )
            return utils.db_provider_to_skosprovider(db_provider)
        except NoResultFound:
            db_provider = provider.create_provider(
                json_data=self.request.openapi_validated.body,
                session=self.request.db,
                skos_registry=self.request.skos_registry,
            )
            self.request.response.status_code = 201
            return utils.db_provider_to_skosprovider(db_provider)

    @view_config(
        route_name='atramhasis.provider',
        permission='delete-provider',
        request_method='DELETE',
    )
    def delete_provider(self):
        provider.delete_provider(
            provider_id=self.request.matchdict['id'],
            session=self.request.db,
        )
        return HTTPNoContent()



import os
import urllib

from pyramid.httpexceptions import HTTPFound
from pyramid.i18n import TranslationStringFactory
from pyramid.response import FileResponse
from pyramid.response import Response
from pyramid.threadlocal import get_current_registry
from pyramid.view import view_config
from pyramid.view import view_defaults
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import LabelType
from skosprovider_sqlalchemy.models import NoteType
from sqlalchemy.exc import NoResultFound

from atramhasis.audit import audit
from atramhasis.cache import invalidate_cache
from atramhasis.cache import invalidate_scheme_cache
from atramhasis.cache import list_region
from atramhasis.cache import tree_region
from atramhasis.errors import ConceptNotFoundException
from atramhasis.errors import ConceptSchemeNotFoundException
from atramhasis.errors import SkosRegistryNotFoundException
from atramhasis.utils import provider_is_external
from atramhasis.utils import update_last_visited_concepts


def labels_to_string(labels, ltype):
    labelstring = ''
    for label in (label for label in labels if label.labeltype_id == ltype):
        labelstring += label.label + ' (' + label.language_id + '), '
    return labelstring[:-2]


def get_definition(notes):
    for note in notes:
        if note.notetype_id == 'definition':
            return note.note


def sort_by_labels(concepts, locale, reverse=False):
    return sorted((x for x in concepts if x.label(locale)),
                  reverse=reverse,
                  key=lambda child: child.label(locale).label.lower()
                  ) + [x for x in concepts if not x.label(locale)]

def get_public_conceptschemes(skos_registry):
    """
    Get all conceptschemes that are visible through the public UI.
    """
    conceptschemes = [
        {'id': x.get_metadata()['id'],
         'conceptscheme': x.concept_scheme}
        for x in skos_registry.get_providers() if not any([not_shown in x.get_metadata()['subject']
                                                           for not_shown in ['external', 'hidden']])
        ]

    return conceptschemes


@view_defaults(accept='text/html', request_method='GET')
class AtramhasisView:
    """
    This object groups HTML views part of the public user interface.
    """

    def __init__(self, request):
        self.request = request
        self.skos_manager = self.request.data_managers['skos_manager']
        self.conceptscheme_manager = self.request.data_managers['conceptscheme_manager']
        if hasattr(request, 'skos_registry') and request.skos_registry is not None:
            self.skos_registry = self.request.skos_registry
        else:
            raise SkosRegistryNotFoundException()

    def _read_request_param(self, param):
        value = None
        if param in self.request.params:
            value = self.request.params.getone(param).strip()
            if not value:
                value = None  # pragma: no cover
        return value

    @view_config(name='favicon.ico')
    def favicon_view(self):
        """
        Return the favicon when requested from the web root.
        """
        here = os.path.dirname(__file__)
        icon = os.path.join(os.path.dirname(here), 'static', 'img', 'favicon.ico')
        response = FileResponse(
            icon,
            request=self.request,
            content_type='image/x-icon'
        )
        return response


    @view_config(route_name='home', renderer='atramhasis:templates/home.jinja2')
    def home_view(self):
        """
        Display the homepage.
        """
        return {'conceptschemes': get_public_conceptschemes(self.skos_registry)}

    @view_config(route_name='skosprovider.conceptschemes', renderer='atramhasis:templates/conceptschemes.jinja2')
    def conceptschemes_view(self):
        """
        Display a list of available conceptschemes.
        """
        return {'conceptschemes': get_public_conceptschemes(self.skos_registry)}

    @audit
    @view_config(route_name='skosprovider.conceptscheme', renderer='atramhasis:templates/conceptscheme.jinja2')
    def conceptscheme_view(self):
        """
        Display a single conceptscheme.
        """
        conceptschemes = get_public_conceptschemes(self.skos_registry)

        scheme_id = self.request.matchdict['scheme_id']

        provider = self.request.skos_registry.get_provider(scheme_id)
        if provider is False:
            raise ConceptSchemeNotFoundException(scheme_id)

        conceptscheme = provider.concept_scheme
        if 'atramhasis.force_display_label_language' in provider.metadata:
            locale = provider.metadata['atramhasis.force_display_label_language']
        else:
            locale = self.request.locale_name
        title = (conceptscheme.label(locale).label if (conceptscheme.label())
                 else scheme_id)

        scheme = {
            'scheme_id': scheme_id,
            'title': title,
            'uri': conceptscheme.uri,
            'labels': conceptscheme.labels,
            'notes': conceptscheme.notes,
            'sources': conceptscheme.sources,
            'top_concepts': provider.get_top_concepts(),
        }

        return {
            'conceptscheme': scheme,
            'conceptschemes': conceptschemes,
            'locale': locale,
            'show_tree': not provider_is_external(provider),
        }

    @audit
    @view_config(route_name='skosprovider.c', renderer='atramhasis:templates/concept.jinja2')
    def concept_view(self):
        """
        Display all about a single concept or collection.
        """
        conceptschemes = [
            {'id': x.get_metadata()['id'],
             'conceptscheme': x.concept_scheme}
            for x in self.skos_registry.get_providers() if not any([not_shown in x.get_metadata()['subject']
                                                                    for not_shown in ['external', 'hidden']])
            ]

        scheme_id = self.request.matchdict['scheme_id']
        c_id = self.request.matchdict['c_id']
        provider = self.request.skos_registry.get_provider(scheme_id)
        label = self._read_request_param('label')
        requested_type = self._read_request_param('type')

        if not provider:
            raise ConceptSchemeNotFoundException(scheme_id)
        if 'atramhasis.force_display_label_language' in provider.metadata:
            locale = provider.metadata['atramhasis.force_display_label_language']
        else:
            locale = self.request.locale_name
        try:
            c = self.skos_manager.get_thing(c_id, provider.conceptscheme_id)
            if isinstance(c, Concept):
                concept_type = "Concept"
            elif isinstance(c, Collection):
                concept_type = "Collection"
            else:
                return Response('Thing without type: ' + str(c_id), status_int=500)
            url = self.request.route_url('skosprovider.c', scheme_id=scheme_id, c_id=c_id)
            update_last_visited_concepts(self.request, {'label': c.label(locale).label, 'url': url})
            return {
                'concept': c, 
                'conceptType': concept_type, 
                'scheme_id': scheme_id,
                'conceptschemes': conceptschemes, 
                'provider': provider,
                'locale': locale,
                'type': requested_type,
                'label': label,
                'show_tree': not provider_is_external(provider)
            }
        except NoResultFound:
            raise ConceptNotFoundException(c_id)

    @view_config(route_name='skosprovider.conceptscheme.cs', renderer='atramhasis:templates/search_result.jinja2')
    def search_result(self):
        """
        Display search results
        """
        conceptschemes = [
            {'id': x.get_metadata()['id'],
             'conceptscheme': x.concept_scheme}
            for x in self.skos_registry.get_providers() if not any([not_shown in x.get_metadata()['subject']
                                                                    for not_shown in ['external', 'hidden']])
            ]

        scheme_id = self.request.matchdict['scheme_id']
        label = self._read_request_param('label')
        requested_type = self._read_request_param('type')
        provider = self.skos_registry.get_provider(scheme_id)
        if provider:
            if 'atramhasis.force_display_label_language' in provider.metadata:
                locale = provider.metadata['atramhasis.force_display_label_language']
            else:
                locale = self.request.locale_name
            if label is not None:
                concepts = provider.find(
                    {'label': label, 'type': requested_type},
                    language=locale,
                    sort='label'
                )
            elif (label is None) and (requested_type is not None):
                concepts = provider.find(
                    {'type': requested_type}, language=locale, sort='label'
                )
            else:
                concepts = provider.get_all(language=locale, sort='label')
            return {
                'concepts': concepts,
                'scheme_id': scheme_id,
                'conceptschemes': conceptschemes,
                'type': requested_type,
                'label': label
            }
        return Response(content_type='text/plain', status_int=404)

    @view_config(route_name='locale')
    def set_locale_cookie(self):
        """
        Set a language cookie to remember what language a user requested.
        """
        settings = get_current_registry().settings
        default_lang = settings.get('pyramid.default_locale_name')
        available_languages = settings.get('available_languages', default_lang).split()
        [x.lower() for x in available_languages]
        language = self.request.GET.get('language', default_lang).lower()
        if language not in available_languages:
            language = default_lang

        referer = self.request.referer
        if referer is not None:
            response = HTTPFound(location=referer)
        else:
            response = HTTPFound(location=self.request.route_url('home'))

        response.set_cookie('_LOCALE_',
                            value=language,
                            max_age=31536000)  # max_age = year
        return response

    @audit
    @view_config(route_name='skosprovider.conceptscheme.csv', renderer='csv')
    def results_csv(self):
        """
        Download search results in CSV format, allowing further processing.
        """
        header = ['conceptscheme', 'id', 'uri', 'type', 'label', 'prefLabels', 'altLabels', 'definition', 'broader',
                  'narrower', 'related']
        rows = []
        scheme_id = self.request.matchdict['scheme_id']
        label = self._read_request_param('label')
        requested_type = self._read_request_param('type')
        provider = self.skos_registry.get_provider(scheme_id)
        if provider:
            if label is not None:
                concepts = self.conceptscheme_manager.find(
                    provider.conceptscheme_id, {'label': label, 'type': requested_type}
                )
            elif (label is None) and (requested_type is not None):
                concepts = self.conceptscheme_manager.find(
                    provider.conceptscheme_id, {'type': requested_type}
                )
            else:
                concepts = self.conceptscheme_manager.get_all(provider.conceptscheme_id)
            for concept in concepts:
                if concept.type == 'concept':
                    rows.append((
                        scheme_id, concept.concept_id, concept.uri, concept.type,
                        concept.label(self.request.locale_name).label,
                        labels_to_string(concept.labels, 'prefLabel'), labels_to_string(concept.labels, 'altLabel'),
                        get_definition(concept.notes), [c.concept_id for c in concept.broader_concepts],
                        [c.concept_id for c in concept.narrower_concepts],
                        [c.concept_id for c in concept.related_concepts]))
                else:
                    rows.append((
                        scheme_id, concept.concept_id, concept.uri, concept.type,
                        concept.label(self.request.locale_name).label,
                        labels_to_string(concept.labels, 'prefLabel'), labels_to_string(concept.labels, 'altLabel'),
                        get_definition(concept.notes), '', [c.concept_id for c in concept.members], ''))
        return {
            'header': header,
            'rows': rows,
            'filename': 'atramhasis_export'
        }

    @view_config(route_name='scheme_tree_html', renderer='scheme_tree.jinja2')
    @view_config(route_name='scheme_tree', renderer='json', accept='application/json')
    def results_tree_json(self):
        scheme_id = self.request.matchdict['scheme_id']
        language = self.request.params.get('language') or self.request.locale_name
        dicts = self.get_results_tree(scheme_id, language)
        if dicts:
            if 'text/html' not in self.request.accept:
                return dicts
            else:
                return {'tree': dicts}
        else:
            return Response(status_int=404)

    @tree_region.cache_on_arguments()
    def get_results_tree(self, scheme_id, locale):
        skostree = self.get_scheme(scheme_id, locale)
        return [self.parse_thing(thing, None) for thing in skostree]

    def get_scheme(self, scheme, locale):
        scheme_tree = []
        provider = self.skos_registry.get_provider(scheme)
        if provider:
            conceptscheme_id = provider.conceptscheme_id

            tco = self.conceptscheme_manager.get_concepts_for_scheme_tree(conceptscheme_id)
            tcl = self.conceptscheme_manager.get_collections_for_scheme_tree(conceptscheme_id)

            scheme_tree = sort_by_labels(tco, locale) + sort_by_labels(tcl, locale)

        return scheme_tree

    def parse_thing(self, thing, parent_tree_id):
        tree_id = self.create_treeid(parent_tree_id, thing.concept_id)
        locale = self.request.params.get('language', self.request.locale_name)

        if thing.type and thing.type == 'collection':
            cs = [member for member in thing.members] if hasattr(thing, 'members') else []
        else:
            cs = [c for c in thing.narrower_concepts]
            cs = cs + [c for c in thing.narrower_collections]

        sortedcs = sort_by_labels(cs, locale)
        children = [self.parse_thing(c, tree_id) for index, c in enumerate(sortedcs, 1)]
        dict_thing = {
            'id': tree_id,
            'concept_id': thing.concept_id,
            'type': thing.type,
            'label': thing.label(locale).label if thing.label(locale) else None,
            'children': children
        }

        return dict_thing

    @staticmethod
    def create_treeid(parent_tree_id, concept_id):
        if parent_tree_id is None:
            return urllib.parse.quote(str(concept_id), safe="")
        else:
            return parent_tree_id + "|" + urllib.parse.quote(str(concept_id), safe="")


@view_defaults(accept='application/json', renderer='json')
class AtramhasisListView:
    """
    This object groups list views part for the user interface.
    """

    def __init__(self, request):
        self.request = request
        self.skos_manager = self.request.data_managers['skos_manager']
        self.localizer = request.localizer
        self._ = TranslationStringFactory('atramhasis')

    @view_config(route_name='labeltypes')
    def labeltype_list_view(self):
        return self.get_list(LabelType)

    @view_config(route_name='notetypes')
    def notetype_list_view(self):
        return self.get_list(NoteType)

    @list_region.cache_on_arguments()
    def get_list(self, listtype):
        return [{"key": ltype.name, "label": self.localizer.translate(self._(ltype.name))}
                for ltype in self.skos_manager.get_by_list_type(listtype)]


@view_defaults(accept='text/html')
class AtramhasisAdminView:
    """
    This object groups HTML views part of the admin user interface.
    """

    def __init__(self, request):
        self.request = request
        self.logged_in = request.authenticated_userid
        if hasattr(request, 'skos_registry') and request.skos_registry is not None:
            self.skos_registry = self.request.skos_registry
        else:
            raise SkosRegistryNotFoundException()

    @view_config(route_name='admin', renderer='atramhasis:templates/admin.jinja2', permission='edit')
    def admin_view(self):
        return {'admin': None}

    @view_config(route_name='scheme_tree_invalidate', renderer='json', accept='application/json', permission='edit')
    def invalidate_scheme_tree(self):
        scheme_id = self.request.matchdict['scheme_id']
        invalidate_scheme_cache(scheme_id)
        return Response(status_int=200)

    @view_config(route_name='tree_invalidate', renderer='json', accept='application/json', permission='edit')
    def invalidate_tree(self):
        invalidate_cache()
        return Response(status_int=200)



"""199 Sources.

Revision ID: 1a4d62b02630
Revises: 2a38d364113b
Create Date: 2015-12-08 12:06:20.303601

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1a4d62b02630"
down_revision = "2a38d364113b"


def upgrade():

    op.create_table(
        "source",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("citation", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "conceptscheme_source",
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.Column("source_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
        ),
        sa.ForeignKeyConstraint(
            ["source_id"],
            ["source.id"],
        ),
        sa.PrimaryKeyConstraint("conceptscheme_id", "source_id"),
    )

    op.create_table(
        "concept_source",
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.Column("source_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["source_id"],
            ["source.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id", "source_id"),
    )


def downgrade():

    op.drop_table("conceptscheme_source")
    op.drop_table("concept_source")
    op.drop_table("source")



"""
Module containing error views.
"""

import logging
import sys

from pyramid.httpexceptions import HTTPMethodNotAllowed
from pyramid.view import notfound_view_config
from pyramid.view import view_config
from pyramid_openapi3 import RequestValidationError
from pyramid_openapi3 import ResponseValidationError
from pyramid_openapi3 import extract_errors
from pyramid_openapi3 import openapi_validation_error
from skosprovider.exceptions import ProviderUnavailableException
from sqlalchemy.exc import IntegrityError

from atramhasis.errors import SkosRegistryNotFoundException
from atramhasis.errors import ValidationError
from atramhasis.protected_resources import ProtectedResourceException


log = logging.getLogger(__name__)


@notfound_view_config(accept='application/json', renderer='json')
@notfound_view_config(accept='text/html', renderer='notfound.jinja2')
def failed_not_found(exc, request):
    """
    View invoked when a resource could not be found.
    """
    log.debug(exc.explanation)
    request.response.status_int = 404
    return {'message': exc.explanation}


@view_config(
    accept='application/json', context=SkosRegistryNotFoundException, renderer='json'
)
@view_config(
    accept='text/html',
    context=SkosRegistryNotFoundException,
    renderer='notfound.jinja2',
)
def failed_skos(exc, request):
    """
    View invoked when Atramhasis can't find a SKOS registry.
    """
    log.error(exc.value, exc_info=sys.exc_info())
    request.response.status_int = 404
    return {'message': exc.value}


@view_config(accept='application/json', context=ValidationError, renderer='json')
def failed_validation(exc, request):
    """
    View invoked when bad data was submitted to Atramhasis.
    """
    log.debug(f"'message': {exc.value}, 'errors': {exc.errors}")
    request.response.status_int = 400
    return {'message': exc.value, 'errors': exc.errors}


@view_config(
    accept='application/json', context=ProtectedResourceException, renderer='json'
)
def protected(exc, request):
    """
    when a protected operation is called on a resource that is still referenced
    """
    log.warning(f"'message': {exc.value}, 'referenced_in': {exc.referenced_in}")
    request.response.status_int = 409
    return {'message': exc.value, 'referenced_in': exc.referenced_in}


@view_config(
    accept='application/json', context=ProviderUnavailableException, renderer='json'
)
def provider_unavailable(exc, request):
    """
    View invoked when ProviderUnavailableException was raised.
    """
    log.error(exc, exc_info=sys.exc_info())
    request.response.status_int = 503
    return {'message': exc.message}


@view_config(accept='application/json', context=IntegrityError, renderer='json')
def data_integrity(exc, request):
    """
    View invoked when IntegrityError was raised.
    """
    log.warning(exc)
    request.response.status_int = 409
    return {
        'message': 'this operation violates the data integrity and could not be executed'
    }


@view_config(accept='application/json', context=Exception, renderer='json')
@view_config(accept='text/html', context=Exception, renderer='error.jinja2')
def failed(exc, request):  # pragma no cover
    """
    View invoked when bad data was submitted to Atramhasis.
    """
    log.error(exc, exc_info=sys.exc_info())
    request.response.status_int = 500
    return {'message': 'unexpected server error'}


@view_config(accept='application/json', context=HTTPMethodNotAllowed, renderer='json')
def failed_method_not_allowed(exc, request):  # pragma no cover
    """
    View invoked when a method is not allowed.
    """
    log.debug(exc.explanation)
    request.response.status_int = 405
    return {'message': exc.explanation}


@view_config(accept='application/json', context=RequestValidationError, renderer='json')
@view_config(
    accept='application/json', context=ResponseValidationError, renderer='json'
)
def failed_openapi_validation(exc, request):
    try:
        errors = [
            f'{error.get("field")}: {error["message"]}'
            for error in list(extract_errors(request, exc.errors))
        ]
        request.response.status_int = 400
        if isinstance(exc, RequestValidationError):
            subject = 'Request'
        else:
            subject = 'Response'
        return {'message': f'{subject} was not valid for schema.', 'errors': errors}
    except Exception as e:
        log.exception('Issue with exception handling.')
        return openapi_validation_error(exc, request)



BLUEBIRDSDATA = [
        {
            'id': 'https://id.bluebirds.org/bird', 
            'uri': 'https://id.bluebirds.org/bird', 
            'type': 'concept', 
            'labels': [{'label': 'Bird', 'type': 'prefLabel', 'language': 'en'}, {'label': 'Vogel', 'type': 'prefLabel', 'language': 'nl'}], 
            'notes': [], 
            'sources': [], 
            'narrower': ['parrot', '579A439C-1A7A-476A-92C3-8A74ABD6B3DB'], 
            'broader': [], 
            'related': [], 
            'member_of': [], 
            'subordinate_arrays': [], 
            'matches': {'close': [], 'exact': [], 'related': [], 'broad': [], 'narrow': []}
        }, 
        {

            'id': 'parrot',
            'uri': 'https://id.bluebirds.org/parrot',
            'type': 'concept',
            'labels': [{'label': 'Parrot', 'type': 'prefLabel', 'language': 'en'}, {'label': 'Papegaai', 'type': 'prefLabel', 'language': 'nl'}],
            'notes': [],
            'sources': [],
            'narrower': ['norwegianblue'],
            'broader': [],
            'related': [],
            'member_of': [],
            'subordinate_arrays': [], 'matches': {'close': [], 'exact': [], 'related': [], 'broad': [], 'narrow': []}
        },
        {
            'id': 'norwegianblue',
            'uri': 'https://id.bluebirds.org/norwegianblue',
            'type': 'concept',
            'labels': [{'label': 'Norwegian Blue', 'type': 'prefLabel', 'language': 'en'}, {'label': 'Noorse Blauwe', 'type': 'prefLabel', 'language': 'nl'}],
            'notes': [],
            'sources': [],
            'narrower': [],
            'broader': ['parrot'],
            'related': [],
            'member_of': [],
            'subordinate_arrays': [],
            'matches': {'close': [], 'exact': [], 'related': ['https://id.slugs.org/slug'], 'broad': [], 'narrow': []}
        },
        {
            'id': '579A439C-1A7A-476A-92C3-8A74ABD6B3DB',
            'uri': 'https://id.bluebirds.org/579A439C-1A7A-476A-92C3-8A74ABD6B3DB',
            'type': 'concept',
            'labels': [{'label': 'Heron', 'type': 'prefLabel', 'language': 'en'}, {'label': 'Héron', 'type': 'prefLabel', 'language': 'fr'}, {'label': 'Reiger', 'type': 'prefLabel', 'language': 'nl'}],
            'notes': [],
            'sources': [],
            'narrower': ['blauwereiger'],
            'broader': [],
            'related': [],
            'member_of': [],
            'subordinate_arrays': [],
            'matches': {'close': [], 'exact': [], 'related': [], 'broad': [], 'narrow': []}
        }, 
        {
            'id': 'blauwereiger',
            'uri': 'https://id.bluebirds.org/blauwereiger',
            'type': 'concept',
            'labels': [{'label': 'Grey heron', 'type': 'prefLabel', 'language': 'en'}, {'label': 'Héron cendré', 'type': 'prefLabel', 'language': 'fr'}, {'label': 'Blauwe reiger', 'type': 'prefLabel', 'language': 'nl'}, {'label': 'Ardea cinerea', 'type': 'prefLabel', 'language': 'la'}],
            'notes': [],
            'sources': [],
            'narrower': [],
            'broader': [],
            'related': [],
            'member_of': [],
            'subordinate_arrays': [],
            'matches': {'close': [], 'exact': [], 'related': [], 'broad': [], 'narrow': []}
        }
]



"""alter ConceptVisitLog concept_id to string

Revision ID: 29306f749043
Revises: b2a7d7614973
Create Date: 2024-07-30 09:12:37.521748

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '29306f749043'
down_revision = 'b2a7d7614973'


def upgrade():
    with op.batch_alter_table('concept_visit_log') as batch_op:
        batch_op.alter_column('concept_id', type_=sa.String, existing_type=sa.Integer)


def downgrade():
    # Drop concept_visit_log which have non-integer concept_id
    op.execute(
        "DELETE FROM concept_visit_log "
        "WHERE cast(cast(concept_id AS INTEGER) AS TEXT) != concept_id"
    )
    with op.batch_alter_table('concept_visit_log') as batch_op:
        batch_op.alter_column('concept_id', type_=sa.Integer, existing_type=sa.String)



import os

from pyramid.config import Configurator
from sqlalchemy import engine_from_config

from atramhasis.data.models import Base


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""

    # set up dump location
    dump_location = settings["atramhasis.dump_location"]
    if not os.path.exists(dump_location):
        os.makedirs(dump_location)

    with Configurator(settings=settings) as config:
        # Set up atramhasis
        config.include("atramhasis")
        # Set up atramhasis db
        config.include("atramhasis:data.db")

        config.scan()
        return config.make_wsgi_app()
